import json
import warnings
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import scipy
import torch
from sklearn.utils import check_array


def create_explain_matrix(
    input_dim: int,
    cat_emb_dim: Union[int, List[int]],
    cat_idxs: List[int],
    post_embed_dim: int,
) -> scipy.sparse.csc_matrix:
    """
    This is a computational trick.
    In order to rapidly sum importances from same embeddings
    to the initial index.

    Parameters
    ----------
    input_dim : int
        Initial input dim
    cat_emb_dim : int or list of int
        if int : size of embedding for all categorical feature
        if list of int : size of embedding for each categorical feature
    cat_idxs : list of int
        Initial position of categorical features
    post_embed_dim : int
        Post embedding inputs dimension

    Returns
    -------
    reducing_matrix : np.array
        Matrix of dim (post_embed_dim, input_dim)  to performe reduce
    """

    if isinstance(cat_emb_dim, int):
        all_emb_impact = [cat_emb_dim - 1] * len(cat_idxs)
    else:
        all_emb_impact = [emb_dim - 1 for emb_dim in cat_emb_dim]

    acc_emb = 0
    nb_emb = 0
    indices_trick = []
    for i in range(input_dim):
        if i not in cat_idxs:
            indices_trick.append([i + acc_emb])
        else:
            indices_trick.append(
                range(i + acc_emb, i + acc_emb + all_emb_impact[nb_emb] + 1)  # type: ignore
            )
            acc_emb += all_emb_impact[nb_emb]
            nb_emb += 1

    reducing_matrix = np.zeros((post_embed_dim, input_dim))
    for i, cols in enumerate(indices_trick):
        reducing_matrix[cols, i] = 1

    return scipy.sparse.csc_matrix(reducing_matrix)


def create_group_matrix(list_groups: List[List[int]], input_dim: int) -> torch.Tensor:
    """
    Create the group matrix corresponding to the given list_groups

    Parameters
    ----------
    - list_groups : list of list of int
        Each element is a list representing features in the same group.
        One feature should appear in maximum one group.
        Feature that don't get assigned a group will be in their own group of one feature.
    - input_dim : number of feature in the initial dataset

    Returns
    -------
    - group_matrix : torch matrix
        A matrix of size (n_groups, input_dim)
        where m_ij represents the importance of feature j in group i
        The rows must some to 1 as each group is equally important a priori.

    """
    check_list_groups(list_groups, input_dim)

    if len(list_groups) == 0:
        group_matrix = torch.eye(input_dim)
        return group_matrix
    else:
        n_groups = input_dim - int(np.sum([len(gp) - 1 for gp in list_groups]))
        group_matrix = torch.zeros((n_groups, input_dim))

        remaining_features = [feat_idx for feat_idx in range(input_dim)]

        current_group_idx = 0
        for group in list_groups:
            group_size = len(group)
            for elem_idx in group:
                # add importrance of element in group matrix and corresponding group
                group_matrix[current_group_idx, elem_idx] = 1 / group_size
                # remove features from list of features
                remaining_features.remove(elem_idx)
            # move to next group
            current_group_idx += 1
        # features not mentionned in list_groups get assigned their own group of singleton
        for remaining_feat_idx in remaining_features:
            group_matrix[current_group_idx, remaining_feat_idx] = 1
            current_group_idx += 1
        return group_matrix


def check_list_groups(list_groups: List[List[int]], input_dim: int) -> None:
    """
    Check that list groups:
        - is a list of list
        - does not contain twice the same feature in different groups
        - does not contain unknown features (>= input_dim)
        - does not contain empty groups
    Parameters
    ----------
    - list_groups : list of list of int
        Each element is a list representing features in the same group.
        One feature should appear in maximum one group.
        Feature that don't get assign a group will be in their own group of one feature.
    - input_dim : number of feature in the initial dataset
    """
    assert isinstance(list_groups, list), "list_groups must be a list of list."

    if len(list_groups) == 0:
        return
    else:
        for group_pos, group in enumerate(list_groups):
            msg = f"Groups must be given as a list of list, but found {group} in position {group_pos}."  # noqa
            assert isinstance(group, list), msg
            assert len(group) > 0, "Empty groups are forbidding please remove empty groups []"

    n_elements_in_groups = np.sum([len(group) for group in list_groups])
    flat_list = []
    for group in list_groups:
        flat_list.extend(group)
    unique_elements = np.unique(flat_list)
    n_unique_elements_in_groups = len(unique_elements)
    msg = "One feature can only appear in one group, please check your grouped_features."
    assert n_unique_elements_in_groups == n_elements_in_groups, msg

    highest_feat = np.max(unique_elements)
    assert highest_feat < input_dim, f"Number of features is {input_dim} but one group contains {highest_feat}."  # noqa
    return


def filter_weights(weights: Union[int, List, np.ndarray]) -> None:
    """
    This function makes sure that weights are in correct format for
    regression and multitask TabNet

    Parameters
    ----------
    weights : int, dict or list
        Initial weights parameters given by user

    Returns
    -------
    None : This function will only throw an error if format is wrong
    """
    err_msg = """Please provide a list or np.array of weights for """
    err_msg += """regression, multitask or pretraining: """
    if isinstance(weights, int):
        if weights == 1:
            raise ValueError(err_msg + "1 given.")
    if isinstance(weights, dict):
        raise ValueError(err_msg + "Dict given.")
    return


def validate_eval_set(
    eval_set: List[Tuple[np.ndarray, np.ndarray]],
    eval_name: Optional[List[str]],
    X_train: np.ndarray,
    y_train: np.ndarray,
) -> Tuple[List[str], List[Tuple[np.ndarray, np.ndarray]]]:
    """Check if the shapes of eval_set are compatible with (X_train, y_train).

    Parameters
    ----------
    eval_set : list of tuple
        List of eval tuple set (X, y).
        The last one is used for early stopping
    eval_name : list of str
        List of eval set names.
    X_train : np.ndarray
        Train owned products
    y_train : np.array
        Train targeted products

    Returns
    -------
    eval_names : list of str
        Validated list of eval_names.
    eval_set : list of tuple
        Validated list of eval_set.

    """
    eval_name = eval_name or [f"val_{i}" for i in range(len(eval_set))]

    assert len(eval_set) == len(eval_name), "eval_set and eval_name have not the same length"
    if len(eval_set) > 0:
        assert all(len(elem) == 2 for elem in eval_set), "Each tuple of eval_set need to have two elements"
    for name, (X, y) in zip(eval_name, eval_set, strict=False):
        check_input(X)
        msg = f"Dimension mismatch between X_{name} " + f"{X.shape} and X_train {X_train.shape}"
        assert len(X.shape) == len(X_train.shape), msg

        msg = f"Dimension mismatch between y_{name} " + f"{y.shape} and y_train {y_train.shape}"
        assert len(y.shape) == len(y_train.shape), msg

        msg = f"Number of columns is different between X_{name} " + f"({X.shape[1]}) and X_train ({X_train.shape[1]})"
        assert X.shape[1] == X_train.shape[1], msg

        if len(y_train.shape) == 2:
            msg = f"Number of columns is different between y_{name} " + f"({y.shape[1]}) and y_train ({y_train.shape[1]})"
            assert y.shape[1] == y_train.shape[1], msg
        msg = f"You need the same number of rows between X_{name} " + f"({X.shape[0]}) and y_{name} ({y.shape[0]})"
        assert X.shape[0] == y.shape[0], msg

    return eval_name, eval_set


def define_device(device_name: str) -> str:
    """
    Define the device to use during training and inference.
    If auto it will detect automatically whether to use cuda or cpu

    Parameters
    ----------
    device_name : str
        Either "auto", "cpu" or "cuda"

    Returns
    -------
    str
        Either "cpu" or "cuda"
    """
    if device_name == "auto":
        if torch.cuda.is_available():
            return "cuda"
        else:
            return "cpu"
    elif device_name == "cuda" and not torch.cuda.is_available():
        return "cpu"
    else:
        return device_name


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj: object) -> object:
        if isinstance(obj, (np.generic, np.ndarray)):
            return obj.tolist()
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def check_input(X: np.ndarray) -> None:
    """
    Raise a clear error if X is a pandas dataframe
    and check array according to scikit rules
    """
    if isinstance(X, (pd.DataFrame, pd.Series)):
        err_message = "Pandas DataFrame are not supported: apply X.values when calling fit"
        raise TypeError(err_message)
    check_array(X, accept_sparse=True)


def check_warm_start(warm_start: bool, from_unsupervised: Optional[bool]) -> None:
    """
    Gives a warning about ambiguous usage of the two parameters.
    """
    if warm_start and from_unsupervised is not None:
        warn_msg = "warm_start=True and from_unsupervised != None: "
        warn_msg = "warm_start will be ignore, training will start from unsupervised weights"
        warnings.warn(warn_msg, stacklevel=2)
    return


def check_embedding_parameters(
    cat_dims: List[int], cat_idxs: List[int], cat_emb_dim: Union[int, List[int]]
) -> Tuple[List[int], List[int], List[int]]:
    """
    Check parameters related to embeddings and rearrange them in a unique manner.
    """
    if (cat_dims == []) ^ (cat_idxs == []):
        if cat_dims == []:
            msg = "If cat_idxs is non-empty, cat_dims must be defined as a list of same length."
        else:
            msg = "If cat_dims is non-empty, cat_idxs must be defined as a list of same length."
        raise ValueError(msg)
    elif len(cat_dims) != len(cat_idxs):
        msg = "The lists cat_dims and cat_idxs must have the same length."
        raise ValueError(msg)

    if isinstance(cat_emb_dim, int):
        cat_emb_dims = [cat_emb_dim] * len(cat_idxs)
    else:
        cat_emb_dims = cat_emb_dim

    # check that all embeddings are provided
    if len(cat_emb_dims) != len(cat_dims):
        msg = f"""cat_emb_dim and cat_dims must be lists of same length, got {len(cat_emb_dims)}
                    and {len(cat_dims)}"""
        raise ValueError(msg)

    # Rearrange to get reproducible seeds with different ordering
    if len(cat_idxs) > 0:
        sorted_idxs = np.argsort(cat_idxs)
        cat_dims = [cat_dims[i] for i in sorted_idxs]
        cat_emb_dims = [cat_emb_dims[i] for i in sorted_idxs]

    return cat_dims, cat_idxs, cat_emb_dims



import math
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import scipy
import torch
from torch.utils.data import DataLoader, Dataset, WeightedRandomSampler

from pytorch_tabnet.utils import check_input

X_type = Union[np.ndarray, scipy.sparse.csr_matrix]
tn_type = Union[torch.Tensor, None]


class TorchDataset(Dataset):
    """
    Format for numpy array

    Parameters
    ----------
    X : 2D array
        The input matrix
    y : 2D array
        The one-hot encoded target
    """

    def __init__(self, x: np.ndarray, y: np.ndarray):
        self.x = torch.from_numpy(x).float()
        self.y = torch.from_numpy(y).float()

    def __len__(self) -> int:
        return len(self.x)

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, torch.Tensor]:
        x, y = self.x[index], self.y[index]
        return x, y


class SparseTorchDataset(Dataset):
    """
    Format for csr_matrix

    Parameters
    ----------
    X : CSR matrix
        The input matrix
    y : 2D array
        The one-hot encoded target
    """

    def __init__(self, x: scipy.sparse.csr_matrix, y: np.ndarray):
        self.x = torch.from_numpy(x.toarray()).float()
        self.y = torch.from_numpy(y).float()

    def __len__(self) -> int:
        return self.x.shape[0]

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, torch.Tensor]:
        x = self.x[index]
        y = self.y[index]
        return x, y


class PredictDataset(Dataset):
    """
    Format for numpy array

    Parameters
    ----------
    X : 2D array
        The input matrix
    """

    def __init__(self, x: Union[X_type, torch.Tensor]):
        if isinstance(x, torch.Tensor):
            self.x = x
        elif scipy.sparse.issparse(x):
            self.x = torch.from_numpy(x.toarray())
        else:
            self.x = torch.from_numpy(x)
        self.x = self.x.float()

    def __len__(self) -> int:
        return len(self.x)

    def __getitem__(self, index: int) -> torch.Tensor:
        x = self.x[index]
        return x


class SparsePredictDataset(Dataset):
    """
    Format for csr_matrix

    Parameters
    ----------
    X : CSR matrix
        The input matrix
    """

    def __init__(self, x: scipy.sparse.csr_matrix):
        self.x = torch.from_numpy(x.toarray()).float()

    def __len__(self) -> int:
        return self.x.shape[0]

    def __getitem__(self, index: int) -> torch.Tensor:
        x = self.x[index]
        return x


@dataclass
class TBDataLoader:
    name: str
    dataset: Union[PredictDataset, TorchDataset]
    batch_size: int
    weights: Optional[torch.Tensor] = None
    pre_training: bool = False
    drop_last: bool = False
    pin_memory: bool = False
    predict: bool = False
    all_at_once: bool = False

    def __iter__(self) -> Iterable[Tuple[torch.Tensor, tn_type, tn_type]]:
        if self.all_at_once:
            if self.pre_training or isinstance(self.dataset, PredictDataset) or isinstance(self.dataset, SparsePredictDataset):
                yield self.dataset.x, None, None
            else:
                yield self.dataset.x, self.dataset.y, None
            return
        ds_len = len(self.dataset)
        perm = None
        if not self.predict:
            perm = torch.randperm(ds_len, pin_memory=self.pin_memory)
        batched_starts = [i for i in range(0, ds_len, self.batch_size)]
        batched_starts += [0] if len(batched_starts) == 0 else []
        for start in batched_starts[: len(self)]:
            if self.predict:
                yield self.make_predict_batch(ds_len, start)
            else:
                yield self.make_train_batch(ds_len, perm, start)

    def make_predict_batch(self, ds_len: int, start: int) -> Tuple[torch.Tensor, tn_type, None]:
        end_at = start + self.batch_size
        if end_at > ds_len:
            end_at = ds_len
        x, y, w = None, None, None
        if self.pre_training or isinstance(self.dataset, PredictDataset) or isinstance(self.dataset, SparsePredictDataset):
            # return self.dataset.x[start:end_at], None, None
            x = self.dataset.x[start:end_at]

        else:
            x, y = self.dataset.x[start:end_at], self.dataset.y[start:end_at]
        w = None if self.weights is None else self.weights[start:end_at]

        return x, y, w

    def make_train_batch(self, ds_len: int, perm: Optional[torch.Tensor], start: int) -> Tuple[torch.Tensor, tn_type, tn_type]:
        end_at = start + self.batch_size
        left_over = None
        if end_at > ds_len:
            left_over = end_at - ds_len
            end_at = ds_len
        indexes = perm[start:end_at]
        x, y, w = None, None, None
        if self.pre_training:
            x = self.dataset.x[indexes]
        else:
            x, y = self.dataset.x[indexes], self.dataset.y[indexes]
        w = self.get_weights(indexes)

        if left_over is not None:
            lo_indexes = perm[:left_over]

            if self.pre_training:
                x = torch.cat((x, self.dataset.x[lo_indexes]))
                w = self.get_weights(lo_indexes)
            else:
                x = torch.cat((x, self.dataset.x[lo_indexes]))
                y = torch.cat((y, self.dataset.y[lo_indexes]))
                w = None if self.weights is None else torch.cat((w, self.get_weights(lo_indexes)))
        return x, y, w

    def __len__(self) -> int:
        res = math.ceil(len(self.dataset) / self.batch_size)
        need_to_drop_last = self.drop_last and not self.predict
        need_to_drop_last = need_to_drop_last and (res > 1)
        res -= need_to_drop_last
        return res

    def get_weights(self, i: torch.Tensor = None) -> Union[torch.Tensor, None]:
        if self.weights is None:
            return None
        if i is None:
            return self.weights
        return self.weights[i]


def create_dataloaders(
    X_train: X_type,
    y_train: np.ndarray,
    eval_set: List[Tuple[X_type, np.ndarray]],
    weights: Union[int, Dict, Iterable],
    batch_size: int,
    num_workers: int,
    drop_last: bool,
    pin_memory: bool,
) -> Tuple[DataLoader, List[DataLoader]]:
    """
    Create dataloaders with or without subsampling depending on weights and balanced.

    Parameters
    ----------
    X_train : np.ndarray
        Training data
    y_train : np.array
        Mapped Training targets
    eval_set : list of tuple
        List of eval tuple set (X, y)
    weights : either 0, 1, dict or iterable
        if 0 (default) : no weights will be applied
        if 1 : classification only, will balanced class with inverse frequency
        if dict : keys are corresponding class values are sample weights
        if iterable : list or np array must be of length equal to nb elements
                      in the training set
    batch_size : int
        how many samples per batch to load
    num_workers : int
        how many subprocesses to use for data loading. 0 means that the data
        will be loaded in the main process
    drop_last : bool
        set to True to drop the last incomplete batch, if the dataset size is not
        divisible by the batch size. If False and the size of dataset is not
        divisible by the batch size, then the last batch will be smaller
    pin_memory : bool
        Whether to pin GPU memory during training

    Returns
    -------
    train_dataloader, valid_dataloader : torch.DataLoader, torch.DataLoader
        Training and validation dataloaders
    """
    _need_shuffle, _sampler = create_sampler(weights, y_train)
    t_weights = None
    # if isinstance(weights, int) and weights == 1:
    #     t_weights = create_class_weights(y_train,)
    if isinstance(weights, np.ndarray):
        t_weights = torch.from_numpy(weights)

    if scipy.sparse.issparse(X_train):
        train_dataloader = TBDataLoader(
            name="train-data",
            dataset=SparseTorchDataset(X_train.astype(np.float32), y_train),
            batch_size=batch_size,
            weights=t_weights,
            drop_last=drop_last,
            pin_memory=pin_memory,
        )
    else:
        train_dataloader = TBDataLoader(
            name="train-data",
            dataset=TorchDataset(X_train.astype(np.float32), y_train),
            batch_size=batch_size,
            weights=t_weights,
            # sampler=sampler,
            # shuffle=need_shuffle,
            # num_workers=num_workers,
            drop_last=drop_last,
            pin_memory=pin_memory,
        )

    valid_dataloaders = []
    for X, y in eval_set:
        v_t_weights = None

        if scipy.sparse.issparse(X):
            valid_dataloaders.append(
                TBDataLoader(
                    name="val-data",
                    dataset=SparseTorchDataset(X.astype(np.float32), y),
                    batch_size=batch_size,
                    weights=v_t_weights,
                    pin_memory=pin_memory,
                    predict=True,
                    # all_at_once=True,
                )
            )
        else:
            valid_dataloaders.append(
                TBDataLoader(
                    name="val-data",
                    dataset=TorchDataset(X.astype(np.float32), y),
                    batch_size=batch_size,
                    weights=v_t_weights,
                    pin_memory=pin_memory,
                    predict=True,
                    # all_at_once=True,
                )
            )

    return train_dataloader, valid_dataloaders


def create_sampler(weights: Union[int, Dict, Iterable], y_train: np.ndarray) -> Tuple[bool, Optional[WeightedRandomSampler]]:
    """
    This creates a sampler from the given weights

    Parameters
    ----------
    weights : either 0, 1, dict or iterable
        if 0 (default) : no weights will be applied
        if 1 : classification only, will balanced class with inverse frequency
        if dict : keys are corresponding class values are sample weights
        if iterable : list or np array must be of length equal to nb elements
                      in the training set
    y_train : np.array
        Training targets
    """
    if isinstance(weights, int):
        if weights == 0:
            need_shuffle = True
            sampler = None
        elif weights == 1:
            need_shuffle = False
            samples_weight = create_class_weights(y_train)
            sampler = WeightedRandomSampler(samples_weight, len(samples_weight))
        else:
            raise ValueError("Weights should be either 0, 1, dictionnary or list.")
    elif isinstance(weights, dict):
        # custom weights per class
        need_shuffle = False
        samples_weight = np.array([weights[t] for t in y_train])
        sampler = WeightedRandomSampler(samples_weight, len(samples_weight))
    else:
        # custom weights
        if len(weights) != len(y_train):  # type: ignore
            raise ValueError("Custom weights should match number of train samples.")
        need_shuffle = False
        samples_weight = np.array(weights)
        sampler = WeightedRandomSampler(samples_weight, len(samples_weight))
    return need_shuffle, sampler


def create_class_weights(y_train: torch.Tensor, base_size: float = 1.0) -> torch.Tensor:
    class_sample_count = np.array([len(np.where(y_train == t)[0]) for t in np.unique(y_train)])
    weights_ = base_size / class_sample_count
    samples_weight = np.zeros(len(y_train))
    for i, t in enumerate(np.unique(y_train)):
        samples_weight[y_train == t] = weights_[i]
    # samples_weight = np.array([weights_[t] for t in y_train])
    samples_weight = torch.from_numpy(samples_weight)
    samples_weight = samples_weight.double()
    return samples_weight


def create_dataloaders_pt(
    X_train: Union[scipy.sparse.csr_matrix, np.ndarray],
    eval_set: List[Union[scipy.sparse.csr_matrix, np.ndarray]],
    weights: Union[int, Dict[Any, Any], Iterable[Any]],
    batch_size: int,
    num_workers: int,
    drop_last: bool,
    pin_memory: bool,
) -> Tuple[DataLoader, List[DataLoader]]:
    """
    Create dataloaders with or without subsampling depending on weights and balanced.

    Parameters
    ----------
    X_train : np.ndarray or scipy.sparse.csr_matrix
        Training data
    eval_set : list of np.array (for Xs and ys) or scipy.sparse.csr_matrix (for Xs)
        List of eval sets
    weights : either 0, 1, dict or iterable
        if 0 (default) : no weights will be applied
        if 1 : classification only, will balanced class with inverse frequency
        if dict : keys are corresponding class values are sample weights
        if iterable : list or np array must be of length equal to nb elements
                      in the training set
    batch_size : int
        how many samples per batch to load
    num_workers : int
        how many subprocesses to use for data loading. 0 means that the data
        will be loaded in the main process
    drop_last : bool
        set to True to drop the last incomplete batch, if the dataset size is not
        divisible by the batch size. If False and the size of dataset is not
        divisible by the batch size, then the last batch will be smaller
    pin_memory : bool
        Whether to pin GPU memory during training

    Returns
    -------
    train_dataloader, valid_dataloader : torch.DataLoader, torch.DataLoader
        Training and validation dataloaders
    """
    _need_shuffle, _sampler = create_sampler(weights, X_train)

    if scipy.sparse.issparse(X_train):
        train_dataloader = TBDataLoader(
            name="train-data",
            dataset=SparsePredictDataset(X_train),
            batch_size=batch_size,
            # sampler=sampler,
            # shuffle=need_shuffle,
            # num_workers=num_workers,
            drop_last=drop_last,
            pin_memory=pin_memory,
            pre_training=True,
        )
    else:
        train_dataloader = TBDataLoader(
            name="train-data",
            dataset=PredictDataset(X_train),
            batch_size=batch_size,
            # sampler=sampler,
            # shuffle=need_shuffle,
            # num_workers=num_workers,
            drop_last=drop_last,
            pin_memory=pin_memory,
            pre_training=True,
        )

    valid_dataloaders = []
    for X in eval_set:
        if scipy.sparse.issparse(X):
            valid_dataloaders.append(
                TBDataLoader(
                    name="val-data",
                    dataset=SparsePredictDataset(X),
                    batch_size=batch_size,
                    # sampler=sampler,
                    # shuffle=need_shuffle,
                    # num_workers=num_workers,
                    drop_last=drop_last,
                    pin_memory=pin_memory,
                    predict=True,
                    # all_at_once=True,
                )
            )
        else:
            valid_dataloaders.append(
                TBDataLoader(
                    name="val-data",
                    dataset=PredictDataset(X),
                    batch_size=batch_size,
                    # sampler=sampler,
                    # shuffle=need_shuffle,
                    # num_workers=num_workers,
                    drop_last=drop_last,
                    pin_memory=pin_memory,
                    predict=True,
                    # all_at_once=True,
                )
            )

    return train_dataloader, valid_dataloaders


def validate_eval_set(eval_set: List[np.ndarray], eval_name: List[str], X_train: np.ndarray) -> List[str]:
    """Check if the shapes of eval_set are compatible with X_train.

    Parameters
    ----------
    eval_set : List of numpy array
        The list evaluation set.
        The last one is used for early stopping
    eval_name : List[str]
        Names for eval sets.
    X_train : np.ndarray
        Train owned products

    Returns
    -------
    eval_names : list of str
        Validated list of eval_names.

    """
    eval_names = eval_name or [f"val_{i}" for i in range(len(eval_set))]
    assert len(eval_set) == len(eval_names), "eval_set and eval_name have not the same length"

    for set_nb, X in enumerate(eval_set):
        check_input(X)
        msg = f"Number of columns is different between eval set {set_nb}" + f"({X.shape[1]}) and X_train ({X_train.shape[1]})"
        assert X.shape[1] == X_train.shape[1], msg
    return eval_names



from dataclasses import dataclass, field
from functools import partial
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import scipy
import torch

# from torch.utils.data import DataLoader
from pytorch_tabnet.abstract_model import TabModel
from pytorch_tabnet.data_handlers import PredictDataset, SparsePredictDataset, TBDataLoader
from pytorch_tabnet.multiclass_utils import check_output_dim, infer_output_dim
from pytorch_tabnet.utils import filter_weights


@dataclass
class TabNetClassifier(TabModel):
    output_dim: int = None
    weight: Any = field(init=False, default=0)

    def __post_init__(self) -> None:
        super(TabNetClassifier, self).__post_init__()
        self._task: str = "classification"
        self._default_loss: Any = partial(
            torch.nn.functional.cross_entropy,
            reduction="none",
        )
        self._default_metric: str = "accuracy"

    def weight_updater(self, weights: Union[bool, Dict[Union[str, int], Any], Any]) -> Union[bool, Dict[Union[str, int], Any]]:
        if isinstance(weights, int):
            return weights  # type: ignore
        elif isinstance(weights, dict):
            return {self.target_mapper[key]: value for key, value in weights.items()}
        else:
            return weights

    def prepare_target(self, y: np.ndarray) -> np.ndarray:
        return np.vectorize(self.target_mapper.get)(y)

    def compute_loss(
        self,
        y_pred: torch.Tensor,
        y_true: torch.Tensor,
        w: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        class_count = None
        if isinstance(self.weight, int) and self.weight == 1:
            _class_num, class_count = y_true.long().unique(return_counts=True)
            class_count[class_count == 0] = 1

        loss = self.loss_fn(y_pred, y_true.long(), weight=1 / class_count if class_count is not None else None)
        if w is not None:
            loss = loss * w
        return loss.mean()

    def update_fit_params(  # type: ignore[override]
        self,
        X_train: np.ndarray,
        y_train: np.ndarray,
        eval_set: List[Tuple[np.ndarray, np.ndarray]],
        weights: Union[bool, Dict[str, Any]],
    ) -> None:
        output_dim: int
        train_labels: List[Any]
        output_dim, train_labels = infer_output_dim(y_train)
        for _X, y in eval_set:
            check_output_dim(train_labels, y)
        self.output_dim: int = output_dim
        self._default_metric = "auc" if self.output_dim == 2 else "accuracy"
        self.classes_: List[Any] = train_labels
        self.target_mapper: Dict[Any, int] = {class_label: index for index, class_label in enumerate(self.classes_)}
        self.preds_mapper: Dict[str, Any] = {str(index): class_label for index, class_label in enumerate(self.classes_)}
        # self.updated_weights: Union[bool, Dict[Union[str, int], Any]] = self.weight_updater(weights)

    def stack_batches(
        self,
        # list_y_true: List[np.ndarray],
        # list_y_score: List[np.ndarray],
        # ) -> Tuple[np.ndarray, np.ndarray]:
        #     y_true: np.ndarray = np.hstack(list_y_true)
        #     y_score: np.ndarray = np.vstack(list_y_score)
        #     y_score = softmax(y_score, axis=1)
        #     return y_true, y_score
        list_y_true: List[torch.Tensor],
        list_y_score: List[torch.Tensor],
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        y_true: torch.Tensor = torch.hstack(list_y_true)
        y_score: torch.Tensor = torch.vstack(list_y_score)
        y_score = torch.nn.Softmax(dim=1)(y_score)
        return y_true, y_score

    def predict_func(self, outputs: np.ndarray) -> np.ndarray:
        outputs = np.argmax(outputs, axis=1)
        return np.vectorize(self.preds_mapper.get)(outputs.astype(str))

    def predict_proba(self, X: Union[torch.Tensor, np.ndarray]) -> np.ndarray:
        self.network.eval()

        if scipy.sparse.issparse(X):
            dataloader: TBDataLoader = TBDataLoader(
                name="predict",
                dataset=SparsePredictDataset(X),
                batch_size=self.batch_size,
                # shuffle=False,
                predict=True,
            )
        else:
            dataloader = TBDataLoader(
                name="predict",
                dataset=PredictDataset(X),
                batch_size=self.batch_size,
                # shuffle=False,
                predict=True,
            )

        results: List[np.ndarray] = []
        with torch.no_grad():
            for _batch_nb, (data, _, _) in enumerate(dataloader):  # type: ignore
                data = data.to(self.device).float()  # type: ignore

                output: torch.Tensor
                _M_loss: torch.Tensor
                output, _M_loss = self.network(data)
                predictions: np.ndarray = (
                    torch.nn.Softmax(dim=1)(output).cpu().detach().numpy()
                )  # todo: replace with pytorch's torch.vstack
                results.append(predictions)
            res: np.ndarray = np.vstack(results)  # todo: replace with pytorch's torch.vstack
        return res


@dataclass
class TabNetRegressor(TabModel):
    output_dim: int = None

    def __post_init__(self) -> None:
        super(TabNetRegressor, self).__post_init__()
        self._task: str = "regression"
        # self._default_loss: Any = torch.nn.functional.mse_loss
        self._default_loss: Any = partial(
            torch.nn.functional.mse_loss,
            reduction="none",
        )
        self._default_metric: str = "mse"

    def prepare_target(self, y: np.ndarray) -> np.ndarray:
        return y

    def compute_loss(
        self,
        y_pred: torch.Tensor,
        y_true: torch.Tensor,
        w: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        loss = self.loss_fn(
            y_pred,
            y_true,
        )
        if len(loss.shape) != 1:
            loss = torch.mean(loss, dim=1)
        if w is not None:
            loss = loss * w
        return loss.mean()

    def update_fit_params(
        self,
        X_train: np.ndarray,
        y_train: np.ndarray,
        eval_set: List[Tuple[np.ndarray, np.ndarray]],
        weights: Union[bool, np.ndarray],
    ) -> None:
        if len(y_train.shape) != 2:
            msg: str = (
                "Targets should be 2D : (n_samples, n_regression) "
                + f"but y_train.shape={y_train.shape} given.\n"
                + "Use reshape(-1, 1) for single regression."
            )
            raise ValueError(msg)
        self.output_dim: int = y_train.shape[1]
        self.preds_mapper: None = None

        self.updated_weights: Union[bool, np.ndarray] = weights
        filter_weights(self.updated_weights)

    def predict_func(self, outputs: np.ndarray) -> np.ndarray:
        return outputs

    def stack_batches(
        self,
        list_y_true: List[torch.Tensor],
        list_y_score: List[torch.Tensor],
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        y_true: torch.Tensor = torch.vstack(list_y_true)
        y_score: torch.Tensor = torch.vstack(list_y_score)
        return y_true, y_score


MultiTabNetRegressor = TabNetRegressor



import copy
import io
import json
import shutil
import warnings
import zipfile
from abc import abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import scipy
import torch
from scipy.sparse import csc_matrix
from sklearn.base import BaseEstimator
from torch.nn.utils import clip_grad_norm_

# from torch.utils.data import DataLoader
from pytorch_tabnet import tab_network
from pytorch_tabnet.callbacks import (
    Callback,
    CallbackContainer,
    EarlyStopping,
    History,
    LRSchedulerCallback,
)
from pytorch_tabnet.data_handlers import PredictDataset, SparsePredictDataset, TBDataLoader, create_dataloaders
from pytorch_tabnet.metrics import MetricContainer, check_metrics
from pytorch_tabnet.utils import (
    ComplexEncoder,
    check_embedding_parameters,
    check_input,
    check_warm_start,
    create_explain_matrix,
    create_group_matrix,
    define_device,
    validate_eval_set,
)


@dataclass
class TabModel(BaseEstimator):
    """Class for TabNet model."""

    compile_backends = ["inductor", "cudagraphs", "ipex", "onnxrt"]

    n_d: int = 8
    n_a: int = 8
    n_steps: int = 3
    gamma: float = 1.3
    cat_idxs: List[int] = field(default_factory=list)
    cat_dims: List[int] = field(default_factory=list)
    cat_emb_dim: Union[int, List[int]] = 1
    n_independent: int = 2
    n_shared: int = 2
    epsilon: float = 1e-15
    momentum: float = 0.02
    lambda_sparse: float = 1e-3
    seed: int = 0
    clip_value: int = 1
    verbose: int = 1
    optimizer_fn: Any = torch.optim.Adam
    optimizer_params: Dict = field(default_factory=lambda: dict(lr=2e-2))
    scheduler_fn: Any = None
    scheduler_params: Dict = field(default_factory=dict)
    mask_type: str = "sparsemax"
    input_dim: int = None
    output_dim: Union[List[int], int] = None
    device_name: str = "auto"
    n_shared_decoder: int = 1
    n_indep_decoder: int = 1
    grouped_features: List[List[int]] = field(default_factory=list)
    _callback_container: CallbackContainer = field(default=None, repr=False, init=False, compare=False)
    compile_backend: str = ""

    def __post_init__(self) -> None:
        # These are default values needed for saving model
        self.batch_size: int = 1024
        self.virtual_batch_size: int = 128

        torch.manual_seed(self.seed)
        # Defining device
        self.device = torch.device(define_device(self.device_name))
        if self.verbose != 0:
            warnings.warn(f"Device used : {self.device}", stacklevel=2)

        # create deep copies of mutable parameters
        self.optimizer_fn = copy.deepcopy(self.optimizer_fn)
        self.scheduler_fn = copy.deepcopy(self.scheduler_fn)

        updated_params = check_embedding_parameters(self.cat_dims, self.cat_idxs, self.cat_emb_dim)
        self.cat_dims, self.cat_idxs, self.cat_emb_dim = updated_params

    def __update__(self, **kwargs: Any) -> None:
        """
        Updates parameters.
        If does not already exists, creates it.
        Otherwise overwrite with warnings.
        """
        update_list = [
            "cat_dims",
            "cat_emb_dim",
            "cat_idxs",
            "input_dim",
            "mask_type",
            "n_a",
            "n_d",
            "n_independent",
            "n_shared",
            "n_steps",
            "grouped_features",
        ]
        for var_name, value in kwargs.items():
            if var_name in update_list:
                try:
                    exec(f"global previous_val; previous_val = self.{var_name}")
                    if previous_val != value:  # type: ignore # noqa
                        wrn_msg = f"Pretraining: {var_name} changed from {previous_val} to {value}"  # type: ignore # noqa
                        warnings.warn(wrn_msg, stacklevel=2)
                        exec(f"self.{var_name} = value")
                except AttributeError:
                    exec(f"self.{var_name} = value")

    def fit(
        self,
        X_train: np.ndarray,
        y_train: np.ndarray,
        eval_set: Union[None, List[Tuple[np.ndarray, np.ndarray]]] = None,
        eval_name: Union[None, List[str]] = None,
        eval_metric: Union[None, List[str]] = None,
        loss_fn: Union[None, Callable] = None,
        weights: Union[int, Dict, np.array] = 0,
        max_epochs: int = 100,
        patience: int = 10,
        batch_size: int = 1024,
        virtual_batch_size: int = None,
        num_workers: int = 0,
        drop_last: bool = True,
        callbacks: Union[None, List] = None,
        pin_memory: bool = True,
        from_unsupervised: Union[None, "TabModel"] = None,
        warm_start: bool = False,
        augmentations: Union[None, Any] = None,
        compute_importance: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Train a neural network stored in self.network
        Using train_dataloader for training data and
        valid_dataloader for validation.

        Parameters
        ----------
        X_train : np.ndarray
            Train set
        y_train : np.array
            Train targets
        eval_set : list of tuple
            List of eval tuple set (X, y).
            The last one is used for early stopping
        eval_name : list of str
            List of eval set names.
        eval_metric : list of str
            List of evaluation metrics.
            The last metric is used for early stopping.
        loss_fn : callable or None
            a PyTorch loss function
        weights : bool or dictionnary
            0 for no balancing
            1 for automated balancing
            dict for custom weights per class
        max_epochs : int
            Maximum number of epochs during training
        patience : int
            Number of consecutive non improving epoch before early stopping
        batch_size : int
            Training batch size
        virtual_batch_size : int
            Batch size for Ghost Batch Normalization (virtual_batch_size < batch_size)
        num_workers : int
            Number of workers used in torch.utils.data.DataLoader
        drop_last : bool
            Whether to drop last batch during training
        callbacks : list of callback function
            List of custom callbacks
        pin_memory: bool
            Whether to set pin_memory to True or False during training
        from_unsupervised: unsupervised trained model
            Use a previously self supervised model as starting weights
        warm_start: bool
            If True, current model parameters are used to start training
        compute_importance : bool
            Whether to compute feature importance
        """
        # update model name

        self.max_epochs: int = max_epochs
        self.patience: int = patience
        self.batch_size = batch_size
        self.virtual_batch_size = virtual_batch_size or batch_size
        self.num_workers: int = num_workers
        self.drop_last: bool = drop_last
        self.input_dim: int = X_train.shape[1]
        self._stop_training: bool = False
        self.pin_memory: bool = pin_memory and (self.device.type != "cpu")
        self.augmentations = augmentations
        self.compute_importance: bool = compute_importance

        if self.augmentations is not None:
            # This ensure reproducibility
            self.augmentations._set_seed()

        eval_set = eval_set if eval_set else []

        if loss_fn is None:
            self.loss_fn = self._default_loss
        else:
            self.loss_fn = loss_fn

        check_input(X_train)
        check_warm_start(warm_start, from_unsupervised)

        self.update_fit_params(
            X_train,
            y_train,
            eval_set,
            weights,
        )

        # Validate and reformat eval set depending on training data
        eval_names, eval_set = validate_eval_set(eval_set, eval_name, X_train, y_train)

        train_dataloader, valid_dataloaders = self._construct_loaders(
            X_train,
            y_train,
            eval_set,
            self.weight_updater(weights=weights),
        )

        if from_unsupervised is not None:
            # Update parameters to match self pretraining
            self.__update__(**from_unsupervised.get_params())

        if not hasattr(self, "network") or not warm_start:
            # model has never been fitted before of warm_start is False
            self._set_network()
        self._update_network_params()
        self._set_metrics(eval_metric, eval_names)
        self._set_optimizer()
        self._set_callbacks(callbacks)

        if from_unsupervised is not None:
            self.load_weights_from_unsupervised(from_unsupervised)
            warnings.warn("Loading weights from unsupervised pretraining", stacklevel=2)
        # Call method on_train_begin for all callbacks
        self._callback_container.on_train_begin()

        # Training loop over epochs
        for epoch_idx in range(self.max_epochs):
            # Call method on_epoch_begin for all callbacks
            self._callback_container.on_epoch_begin(epoch_idx)
            self._train_epoch(train_dataloader)
            # Apply predict epoch to all eval sets
            for eval_name_, valid_dataloader in zip(eval_names, valid_dataloaders, strict=False):
                self._predict_epoch(eval_name_, valid_dataloader)
            # Call method on_epoch_end for all callbacks
            self._callback_container.on_epoch_end(epoch_idx, logs=self.history.epoch_metrics)

            if self._stop_training:
                break

        # Call method on_train_end for all callbacks
        self._callback_container.on_train_end()
        self.network.eval()

        if self.compute_importance:
            # compute feature importance once the best model is defined
            self.feature_importances_ = self._compute_feature_importances(X_train)

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Make predictions on a batch (valid)

        Parameters
        ----------
        X : a :tensor: `torch.Tensor` or matrix: `scipy.sparse.csr_matrix`
            Input data

        Returns
        -------
        predictions : np.array
            Predictions of the regression problem
        """
        self.network.eval()

        if scipy.sparse.issparse(X):
            dataloader = TBDataLoader(
                name="predict",
                dataset=SparsePredictDataset(X),
                batch_size=self.batch_size,
                # shuffle=False,
                predict=True,
            )
        else:
            dataloader = TBDataLoader(
                name="predict",
                dataset=PredictDataset(X),
                batch_size=self.batch_size,
                # shuffle=False,
                predict=True,
            )

        results = []
        with torch.no_grad():
            for _batch_nb, (data, _, _) in enumerate(iter(dataloader)):  # type: ignore
                data = data.to(self.device, non_blocking=True).float()
                output, _M_loss = self.network(data)
                predictions = output.cpu().detach().numpy()
                results.append(predictions)
        res = np.vstack(results)
        return self.predict_func(res)

    def explain(self, X: np.ndarray, normalize: bool = False) -> Tuple[np.ndarray, Dict]:
        """
        Return local explanation

        Parameters
        ----------
        X : tensor: `torch.Tensor` or matrix: `scipy.sparse.csr_matrix`
            Input data
        normalize : bool (default False)
            Wheter to normalize so that sum of features are equal to 1

        Returns
        -------
        M_explain : matrix
            Importance per sample, per columns.
        masks : matrix
            Sparse matrix showing attention masks used by network.
        """
        self.network.eval()

        if scipy.sparse.issparse(X):
            dataloader = TBDataLoader(
                name="predict",
                dataset=SparsePredictDataset(X),
                batch_size=self.batch_size,
                # shuffle=False,
                predict=True,
            )
        else:
            dataloader = TBDataLoader(
                name="predict",
                dataset=PredictDataset(X),
                batch_size=self.batch_size,
                # shuffle=False,
                predict=True,
            )

        res_explain = []
        with torch.no_grad():
            for batch_nb, (data, _, _) in enumerate(dataloader):  # type: ignore
                data = data.to(self.device, non_blocking=True).float()  # type: ignore

                M_explain, masks = self.network.forward_masks(data)
                for key, value in masks.items():
                    masks[key] = csc_matrix.dot(value.cpu().detach().numpy(), self.reducing_matrix)
                original_feat_explain = csc_matrix.dot(M_explain.cpu().detach().numpy(), self.reducing_matrix)
                res_explain.append(original_feat_explain)

                if batch_nb == 0:
                    res_masks = masks
                else:
                    for key, value in masks.items():
                        res_masks[key] = np.vstack([res_masks[key], value])

        res_explain = np.vstack(res_explain)

        if normalize:
            res_explain /= np.sum(res_explain, axis=1)[:, None]

        return res_explain, res_masks

    def load_weights_from_unsupervised(self, unsupervised_model: "TabModel") -> None:
        update_state_dict = copy.deepcopy(self.network.state_dict())
        for param, weights in unsupervised_model.network.state_dict().items():
            if param.startswith("encoder"):
                # Convert encoder's layers name to match
                new_param = "tabnet." + param
            else:
                new_param = param
            if self.network.state_dict().get(new_param) is not None:
                # update only common layers
                update_state_dict[new_param] = weights

        self.network.load_state_dict(update_state_dict)

    def load_class_attrs(self, class_attrs: Dict) -> None:
        for attr_name, attr_value in class_attrs.items():
            setattr(self, attr_name, attr_value)

    def save_model(self, path: str) -> str:
        """Saving TabNet model in two distinct files.

        Parameters
        ----------
        path : str
            Path of the model.

        Returns
        -------
        str
            input filepath with ".zip" appended

        """
        saved_params = {}
        init_params = {}
        for key, val in self.get_params().items():
            if isinstance(val, type):
                # Don't save torch specific params
                continue
            else:
                init_params[key] = val
        saved_params["init_params"] = init_params

        class_attrs = {"preds_mapper": self.preds_mapper}
        saved_params["class_attrs"] = class_attrs

        # Create folder
        Path(path).mkdir(parents=True, exist_ok=True)

        # Save models params
        with open(Path(path).joinpath("model_params.json"), "w", encoding="utf8") as f:
            json.dump(saved_params, f, cls=ComplexEncoder)

        # Save state_dict
        torch.save(self.network.state_dict(), Path(path).joinpath("network.pt"))
        shutil.make_archive(path, "zip", path)
        shutil.rmtree(path)
        print(f"Successfully saved model at {path}.zip")
        return f"{path}.zip"

    def load_model(self, filepath: str) -> None:
        """Load TabNet model.

        Parameters
        ----------
        filepath : str
            Path of the model.
        """
        try:
            with zipfile.ZipFile(filepath) as z:
                with z.open("model_params.json") as f:
                    loaded_params = json.load(f)
                    loaded_params["init_params"]["device_name"] = self.device_name
                with z.open("network.pt") as f:
                    try:
                        saved_state_dict = torch.load(f, map_location=self.device)
                    except io.UnsupportedOperation:
                        # In Python <3.7, the returned file object is not seekable (which at least
                        # some versions of PyTorch require) - so we'll try buffering it in to a
                        # BytesIO instead:
                        saved_state_dict = torch.load(
                            io.BytesIO(f.read()),
                            map_location=self.device,
                        )
        except KeyError:
            raise KeyError("Your zip file is missing at least one component")

        self.__init__(**loaded_params["init_params"])  # type: ignore

        self._set_network()
        self.network.load_state_dict(saved_state_dict)
        self.network.eval()
        self.load_class_attrs(loaded_params["class_attrs"])

        return

    def _train_epoch(self, train_loader: TBDataLoader) -> None:
        """
        Trains one epoch of the network in self.network

        Parameters
        ----------
        train_loader : a :class: `torch.utils.data.Dataloader`
            DataLoader with train set
        """
        self.network.train()

        for batch_idx, (X, y, w) in enumerate(train_loader):  # type: ignore
            self._callback_container.on_batch_begin(batch_idx)
            X = X.to(  # type: ignore
                self.device,
            )
            y = y.to(  # type: ignore
                self.device,
            )
            if w is not None:  # type: ignore
                w = w.to(  # type: ignore
                    self.device,
                )

            batch_logs = self._train_batch(X, y, w)

            self._callback_container.on_batch_end(batch_idx, batch_logs)

        epoch_logs = {"lr": self._optimizer.param_groups[-1]["lr"]}
        self.history.epoch_metrics.update(epoch_logs)

        return

    def _train_batch(self, X: torch.Tensor, y: torch.Tensor, w: Optional[torch.Tensor] = None) -> Dict:
        """
        Trains one batch of data

        Parameters
        ----------
        X : torch.Tensor
            Train matrix
        y : torch.Tensor
            Target matrix

        Returns
        -------
        batch_outs : dict
            Dictionnary with "y": target and "score": prediction scores.
        batch_logs : dict
            Dictionnary with "batch_size" and "loss".
        """
        batch_logs = {"batch_size": X.shape[0]}

        if self.augmentations is not None:
            X, y = self.augmentations(X, y)

        for param in self.network.parameters():
            param.grad = None

        output, M_loss = self.network(X)

        loss = self.compute_loss(output, y, w)
        # Add the overall sparsity loss
        loss = loss - self.lambda_sparse * M_loss

        # Perform backward pass and optimization
        loss.backward()
        if self.clip_value:
            clip_grad_norm_(self.network.parameters(), self.clip_value)
        self._optimizer.step()

        batch_logs["loss"] = loss.item()

        return batch_logs

    def _predict_epoch(self, name: str, loader: TBDataLoader) -> None:  # todo: replace loader
        """
        Predict an epoch and update metrics.

        Parameters
        ----------
        name : str
            Name of the validation set
        loader : torch.utils.data.Dataloader
                DataLoader with validation set
        """
        # Setting network on evaluation mode
        self.network.eval()

        list_y_true = []
        list_y_score = []
        list_w_ture = []
        with torch.no_grad():
            # Main loop
            for _batch_idx, (X, y, w) in enumerate(loader):  # type: ignore
                scores = self._predict_batch(X.to(self.device, non_blocking=True).float())  # type: ignore
                list_y_true.append(y.to(self.device, non_blocking=True))  # type: ignore

                list_y_score.append(scores)  # todo: weighted scores
                if w is not None:  # type: ignore
                    list_w_ture.append(w.to(self.device, non_blocking=True))  # type: ignore
        w_true = None
        if list_w_ture:
            w_true = torch.cat(list_w_ture, dim=0)

        y_true, scores = self.stack_batches(list_y_true, list_y_score)

        metrics_logs = self._metric_container_dict[name](y_true, scores, w_true)
        self.network.train()
        self.history.epoch_metrics.update(metrics_logs)
        return

    def _predict_batch(self, X: torch.Tensor) -> torch.Tensor:
        """
        Predict one batch of data.

        Parameters
        ----------
        X : torch.Tensor
            Owned products

        Returns
        -------
        np.array
            model scores
        """

        # compute model output
        scores, _ = self.network(X)

        if isinstance(scores, list):
            scores = [x for x in scores]
        else:
            scores = scores

        return scores

    def _set_network(self) -> None:
        """Setup the network and explain matrix."""
        torch.manual_seed(self.seed)

        self.group_matrix = create_group_matrix(self.grouped_features, self.input_dim)

        self.network = tab_network.TabNet(
            self.input_dim,
            self.output_dim,
            n_d=self.n_d,
            n_a=self.n_a,
            n_steps=self.n_steps,
            gamma=self.gamma,
            cat_idxs=self.cat_idxs,
            cat_dims=self.cat_dims,
            cat_emb_dim=self.cat_emb_dim,
            n_independent=self.n_independent,
            n_shared=self.n_shared,
            epsilon=self.epsilon,
            virtual_batch_size=self.virtual_batch_size,
            momentum=self.momentum,
            mask_type=self.mask_type,
            group_attention_matrix=self.group_matrix.to(self.device),
        ).to(self.device)
        if self.compile_backend in self.compile_backends:
            self.network = torch.compile(self.network, backend=self.compile_backend)

        self.reducing_matrix = create_explain_matrix(
            self.network.input_dim,
            self.network.cat_emb_dim,
            self.network.cat_idxs,
            self.network.post_embed_dim,
        )

    def _set_metrics(self, metrics: Union[None, List[str]], eval_names: List[str]) -> None:
        """Set attributes relative to the metrics.

        Parameters
        ----------
        metrics : list of str
            List of eval metric names.
        eval_names : list of str
            List of eval set names.

        """
        metrics = metrics or [self._default_metric]

        metrics = check_metrics(metrics)
        # Set metric container for each sets
        self._metric_container_dict: Dict[str, MetricContainer] = {}
        for name in eval_names:
            self._metric_container_dict.update({name: MetricContainer(metrics, prefix=f"{name}_")})

        self._metrics: List = []
        self._metrics_names: List[str] = []
        for _, metric_container in self._metric_container_dict.items():
            self._metrics.extend(metric_container.metrics)
            self._metrics_names.extend(metric_container.names)

        # Early stopping metric is the last eval metric
        self.early_stopping_metric: Union[None, str] = self._metrics_names[-1] if len(self._metrics_names) > 0 else None

    def _set_callbacks(self, custom_callbacks: Union[None, List]) -> None:
        """Setup the callbacks functions.

        Parameters
        ----------
        custom_callbacks : list of func
            List of callback functions.

        """
        # Setup default callbacks history, early stopping and scheduler
        callbacks: List[Union[History, EarlyStopping, LRSchedulerCallback, Callback]] = []
        self.history = History(self, verbose=self.verbose)
        callbacks.append(self.history)
        if (self.early_stopping_metric is not None) and (self.patience > 0):
            early_stopping = EarlyStopping(
                early_stopping_metric=self.early_stopping_metric,
                is_maximize=(self._metrics[-1]._maximize if len(self._metrics) > 0 else None),
                patience=self.patience,
            )
            callbacks.append(early_stopping)
        else:
            wrn_msg = "No early stopping will be performed, last training weights will be used."
            warnings.warn(wrn_msg, stacklevel=2)

        if self.scheduler_fn is not None:
            # Add LR Scheduler call_back
            is_batch_level = self.scheduler_params.pop("is_batch_level", False)
            scheduler = LRSchedulerCallback(
                scheduler_fn=self.scheduler_fn,
                scheduler_params=self.scheduler_params,
                optimizer=self._optimizer,
                early_stopping_metric=self.early_stopping_metric,
                is_batch_level=is_batch_level,
            )
            callbacks.append(scheduler)

        if custom_callbacks:
            callbacks.extend(custom_callbacks)
        self._callback_container = CallbackContainer(callbacks)
        self._callback_container.set_trainer(self)

    def _set_optimizer(self) -> None:
        """Setup optimizer."""
        self._optimizer = self.optimizer_fn(self.network.parameters(), **self.optimizer_params)

    def _construct_loaders(
        self,
        X_train: np.ndarray,
        y_train: np.ndarray,
        eval_set: List[Tuple[np.ndarray, np.ndarray]],
        weights: Union[int, Dict, np.array],
    ) -> Tuple[TBDataLoader, List[TBDataLoader]]:
        """Generate dataloaders for train and eval set.

        Parameters
        ----------
        X_train : np.array
            Train set.
        y_train : np.array
            Train targets.
        eval_set : list of tuple
            List of eval tuple set (X, y).

        Returns
        -------
        train_dataloader : `torch.utils.data.Dataloader`
            Training dataloader.
        valid_dataloaders : list of `torch.utils.data.Dataloader`
            List of validation dataloaders.

        """
        # all weights are not allowed for this type of model
        y_train_mapped = self.prepare_target(y_train)
        for i, (X, y) in enumerate(eval_set):
            y_mapped = self.prepare_target(y)
            eval_set[i] = (X, y_mapped)

        train_dataloader, valid_dataloaders = create_dataloaders(
            X_train,
            y_train_mapped,
            eval_set,
            weights,
            self.batch_size,
            self.num_workers,
            self.drop_last,
            self.pin_memory,
        )
        return train_dataloader, valid_dataloaders

    def _compute_feature_importances(self, X: np.ndarray) -> np.ndarray:
        """Compute global feature importance.

        Parameters
        ----------
        loader : `torch.utils.data.Dataloader`
            Pytorch dataloader.

        """
        M_explain, _ = self.explain(X, normalize=False)
        sum_explain = M_explain.sum(axis=0)
        feature_importances_ = sum_explain / np.sum(sum_explain)
        return feature_importances_

    def _update_network_params(self) -> None:
        self.network.virtual_batch_size = self.virtual_batch_size

    def weight_updater(self, weights: Union[bool, Dict[Union[str, int], Any], Any]) -> Union[bool, Dict[Union[str, int], Any]]:
        return weights

    @abstractmethod
    def update_fit_params(
        self,
        X_train: np.ndarray,
        y_train: np.ndarray,
        eval_set: List[Tuple[np.ndarray, np.ndarray]],
        weights: Union[int, Dict],
    ) -> None:
        """
        Set attributes relative to fit function.

        Parameters
        ----------
        X_train : np.ndarray
            Train set
        y_train : np.array
            Train targets
        eval_set : list of tuple
            List of eval tuple set (X, y).
        weights : bool or dictionnary
            0 for no balancing
            1 for automated balancing
        """
        raise NotImplementedError("users must define update_fit_params to use this base class")

    @abstractmethod
    def compute_loss(self, *args: Any, **kwargs: Any) -> torch.Tensor:
        """
        Compute the loss.

        Parameters
        ----------
        y_score : a :tensor: `torch.Tensor`
            Score matrix
        y_true : a :tensor: `torch.Tensor`
            Target matrix

        Returns
        -------
        float
            Loss value
        """
        ...

    @abstractmethod
    def prepare_target(self, y: np.ndarray) -> torch.Tensor:
        """
        Prepare target before training.

        Parameters
        ----------
        y : a :tensor: `torch.Tensor`
            Target matrix.

        Returns
        -------
        `torch.Tensor`
            Converted target matrix.
        """
        ...

    @abstractmethod
    def predict_func(self, y_score: np.ndarray) -> np.ndarray: ...

    def stack_batches(self, *args: Any, **kwargs: Any) -> Tuple[torch.Tensor, torch.Tensor]: ...



from dataclasses import dataclass
from typing import Any, List, Union

import torch
from torch.nn import CrossEntropyLoss
from torcheval.metrics.functional import (
    multiclass_accuracy,
    multiclass_auroc,
)


def UnsupervisedLoss(
    y_pred: torch.Tensor,
    embedded_x: torch.Tensor,
    obf_vars: torch.Tensor,
    eps: float = 1e-9,
    weights: torch.Tensor = None,
) -> torch.Tensor:
    """
    Implements unsupervised loss function with optional sample weights.
    """
    errors = y_pred - embedded_x
    reconstruction_errors = torch.mul(errors, obf_vars) ** 2
    batch_means = torch.mean(embedded_x, dim=0)
    batch_means[batch_means == 0] = 1

    batch_stds = torch.std(embedded_x, dim=0) ** 2
    batch_stds[batch_stds == 0] = batch_means[batch_stds == 0]
    features_loss = torch.matmul(reconstruction_errors, 1 / batch_stds)
    nb_reconstructed_variables = torch.sum(obf_vars, dim=1)
    features_loss = features_loss / (nb_reconstructed_variables + eps)

    if weights is not None:
        features_loss = features_loss * weights

    loss = torch.mean(features_loss)
    return loss


@dataclass
class UnsupMetricContainer:
    """Updated to support weights."""

    metric_names: List[str]
    prefix: str = ""

    def __post_init__(self) -> None:
        self.metrics = Metric.get_metrics_by_names(self.metric_names)
        self.names = [self.prefix + name for name in self.metric_names]

    def __call__(
        self,
        y_pred: torch.Tensor,
        embedded_x: torch.Tensor,
        obf_vars: torch.Tensor,
        weights: torch.Tensor = None,
    ) -> dict:
        logs = {}
        for metric in self.metrics:
            res = metric(y_pred, embedded_x, obf_vars, weights)
            logs[self.prefix + metric._name] = res
        return logs


@dataclass
class MetricContainer:
    """Updated to support weights."""

    metric_names: List[str]
    prefix: str = ""

    def __post_init__(self) -> None:
        self.metrics = Metric.get_metrics_by_names(self.metric_names)
        self.names = [self.prefix + name for name in self.metric_names]

    def __call__(
        self,
        y_true: torch.Tensor,
        y_pred: torch.Tensor,
        weights: torch.Tensor = None,
    ) -> dict:
        logs = {}
        for metric in self.metrics:
            if isinstance(y_pred, list):
                res = torch.mean(torch.tensor([metric(y_true[:, i], y_pred[i], weights) for i in range(len(y_pred))]))
            else:
                res = metric(y_true, y_pred, weights)
            logs[self.prefix + metric._name] = res
        return logs


class Metric:
    _name: str
    _maximize: bool

    def __call__(
        self,
        y_true: torch.Tensor,
        y_pred: torch.Tensor,
        weights: torch.Tensor = None,
    ) -> float:
        raise NotImplementedError("Custom Metrics must implement this function")

    @classmethod
    def get_metrics_by_names(cls, names: List[str]) -> List:
        """Get list of metric classes.

        Parameters
        ----------
        cls : Metric
            Metric class.
        names : list
            List of metric names.

        Returns
        -------
        metrics : list
            List of metric classes.

        """
        available_metrics = cls.__subclasses__()
        available_names = [metric()._name for metric in available_metrics]
        metrics = []
        for name in names:
            assert name in available_names, f"{name} is not available, choose in {available_names}"
            idx = available_names.index(name)
            metric = available_metrics[idx]()
            metrics.append(metric)
        return metrics


class AUC(Metric):
    """
    AUC.
    """

    _name: str = "auc"
    _maximize: bool = True

    def __call__(
        self,
        y_true: torch.Tensor,
        y_score: torch.Tensor,
        weights: torch.Tensor = None,
    ) -> float:
        num_of_classes = y_score.shape[1]
        # if weights is not None:
        #     weights = weights.to(y_true.device)
        #     return multiclass_auroc(y_score, y_true, num_classes=num_of_classes, weights=weights).cpu().item()
        return multiclass_auroc(y_score, y_true, num_classes=num_of_classes, average="macro").cpu().item()


class Accuracy(Metric):
    """
    Accuracy.
    """

    _name: str = "accuracy"
    _maximize: bool = True

    def __call__(
        self,
        y_true: torch.Tensor,
        y_score: torch.Tensor,
        weights: torch.Tensor = None,
    ) -> float:
        res = multiclass_accuracy(
            y_score,
            y_true,
        )
        # if weights is not None:
        #     weights = weights.to(y_true.device)
        #     res *= weights
        return res.cpu().item()


class BalancedAccuracy(Metric):
    """
    Balanced Accuracy.
    """

    _name: str = "balanced_accuracy"
    _maximize: bool = True

    def __call__(
        self,
        y_true: torch.Tensor,
        y_score: torch.Tensor,
        weights: torch.Tensor = None,
    ) -> float:
        num_of_classes = y_score.shape[1]
        # if weights is not None:
        #     weights = weights.to(y_true.device)
        #     return multiclass_accuracy(y_score, y_true, average="macro", num_classes=num_of_classes, weights=weights).cpu().item()
        return multiclass_accuracy(y_score, y_true, average="macro", num_classes=num_of_classes).cpu().item()


class LogLoss(Metric):
    """
    LogLoss.
    """

    _name: str = "logloss"
    _maximize: bool = False

    def __call__(
        self,
        y_true: torch.Tensor,
        y_score: torch.Tensor,
        weights: torch.Tensor = None,
    ) -> float:
        loss = CrossEntropyLoss(reduction="none")(y_score.float(), y_true.long())
        if weights is not None:
            loss *= weights.to(y_true.device)
        return loss.mean().item()


class MAE(Metric):
    """
    Mean Absolute Error.
    """

    _name: str = "mae"
    _maximize: bool = False

    def __call__(
        self,
        y_true: torch.Tensor,
        y_score: torch.Tensor,
        weights: torch.Tensor = None,
    ) -> float:
        errors = torch.abs(y_true - y_score)
        if weights is not None:
            errors *= weights.to(y_true.device)
        return torch.mean(errors).cpu().item()


class MSE(Metric):
    """
    Mean Squared Error.
    """

    _name: str = "mse"
    _maximize: bool = False

    def __call__(
        self,
        y_true: torch.Tensor,
        y_score: torch.Tensor,
        weights: torch.Tensor = None,
    ) -> float:
        errors = (y_true - y_score) ** 2
        if weights is not None:
            errors *= weights.to(y_true.device)
        return torch.mean(errors).cpu().item()


class RMSLE(Metric):
    """
    Root Mean squared logarithmic error regression loss.
    Scikit-implementation:
    https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_log_error.html
    Note: In order to avoid error, negative predictions are clipped to 0.
    This means that you should clip negative predictions manually after calling predict.
    """

    _name: str = "rmsle"
    _maximize: bool = False

    def __call__(
        self,
        y_true: torch.Tensor,
        y_score: torch.Tensor,
        weights: torch.Tensor = None,
    ) -> float:
        logerror = torch.log(y_score + 1) - torch.log(y_true + 1)
        squared_logerror = logerror**2
        if weights is not None:
            squared_logerror *= weights.to(y_true.device)
        return torch.sqrt(torch.mean(squared_logerror)).cpu().item()


class UnsupervisedMetric(Metric):
    """
    Unsupervised metric
    """

    _name: str = "unsup_loss"
    _maximize: bool = False

    def __call__(  # type: ignore
        self,
        y_pred: torch.Tensor,
        embedded_x: torch.Tensor,
        obf_vars: torch.Tensor,
        weights: torch.Tensor = None,
    ) -> float:
        loss = UnsupervisedLoss(y_pred, embedded_x, obf_vars, weights=weights)
        return loss.cpu().item()


class UnsupervisedNumpyMetric(Metric):
    """
    Unsupervised metric
    """

    _name: str = "unsup_loss_numpy"
    _maximize: bool = False

    def __call__(  # type: ignore[override]
        # self, y_pred: np.ndarray, embedded_x: np.ndarray, obf_vars: np.ndarray
        self,
        y_pred: torch.Tensor,
        embedded_x: torch.Tensor,
        obf_vars: torch.Tensor,
        weights: torch.Tensor = None,
    ) -> float:
        return UnsupervisedLoss(y_pred, embedded_x, obf_vars).cpu().item()


class RMSE(Metric):
    """
    Root Mean Squared Error.
    """

    _name: str = "rmse"
    _maximize: bool = False

    def __call__(
        self,
        y_true: torch.Tensor,
        y_score: torch.Tensor,
        weights: torch.Tensor = None,
    ) -> float:
        mse_errors = (y_true - y_score) ** 2
        if weights is not None:
            mse_errors *= weights.to(y_true.device)
        return torch.sqrt(torch.mean(mse_errors)).cpu().item()


def check_metrics(metrics: List[Union[str, Any]]) -> List[str]:
    """Check if custom metrics are provided.

    Parameters
    ----------
    metrics : list of str or classes
        List with built-in metrics (str) or custom metrics (classes).

    Returns
    -------
    val_metrics : list of str
        List of metric names.

    """
    val_metrics = []
    for metric in metrics:
        if isinstance(metric, str):
            val_metrics.append(metric)
        elif issubclass(metric, Metric):
            val_metrics.append(metric()._name)
        else:
            raise TypeError("You need to provide a valid metric format")
    return val_metrics



import copy
import datetime
import time
import warnings
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import numpy as np
import torch


class Callback:
    """
    Abstract base class used to build new callbacks.
    """

    def __init__(self) -> None:
        pass

    def set_params(self, params: Dict[str, Any]) -> None:
        self.params = params

    def set_trainer(self, model: Any) -> None:
        self.trainer = model

    def on_epoch_begin(self, epoch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        pass

    def on_epoch_end(self, epoch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        pass

    def on_batch_begin(self, batch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        pass

    def on_batch_end(self, batch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        pass

    def on_train_begin(self, logs: Optional[Dict[str, Any]] = None) -> None:
        pass

    def on_train_end(self, logs: Optional[Dict[str, Any]] = None) -> None:
        pass


@dataclass
class CallbackContainer:
    """
    Container holding a list of callbacks.
    """

    callbacks: List[Callback] = field(default_factory=list)

    def append(self, callback: Callback) -> None:
        self.callbacks.append(callback)

    def set_params(self, params: Dict[str, Any]) -> None:
        for callback in self.callbacks:
            callback.set_params(params)

    def set_trainer(self, trainer: Any) -> None:
        self.trainer = trainer
        for callback in self.callbacks:
            callback.set_trainer(trainer)

    def on_epoch_begin(self, epoch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_epoch_begin(epoch, logs)

    def on_epoch_end(self, epoch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_epoch_end(epoch, logs)

    def on_batch_begin(self, batch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_batch_begin(batch, logs)

    def on_batch_end(self, batch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_batch_end(batch, logs)

    def on_train_begin(self, logs: Optional[Dict[str, Any]] = None) -> None:
        logs = logs or {}
        logs["start_time"] = time.time()
        for callback in self.callbacks:
            callback.on_train_begin(logs)

    def on_train_end(self, logs: Optional[Dict[str, Any]] = None) -> None:
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_train_end(logs)


@dataclass
class EarlyStopping(Callback):
    """EarlyStopping callback to exit the training loop if early_stopping_metric
    does not improve by a certain amount for a certain
    number of epochs.

    Parameters
    ---------
    early_stopping_metric : str
        Early stopping metric name
    is_maximize : bool
        Whether to maximize or not early_stopping_metric
    tol : float
        minimum change in monitored value to qualify as improvement.
        This number should be positive.
    patience : integer
        number of epochs to wait for improvement before terminating.
        the counter be reset after each improvement

    """

    early_stopping_metric: str
    is_maximize: bool
    tol: float = 0.0
    patience: int = 5

    def __post_init__(self) -> None:
        self.best_epoch: int = 0
        self.stopped_epoch: int = 0
        self.wait: int = 0
        self.best_weights: Optional[Any] = None
        self.best_loss: float = np.inf
        if self.is_maximize:
            self.best_loss = -self.best_loss
        super().__init__()

    def on_epoch_end(self, epoch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        current_loss = logs.get(self.early_stopping_metric)
        if current_loss is None:
            return

        loss_change = current_loss - self.best_loss
        max_improved = self.is_maximize and loss_change > self.tol
        min_improved = (not self.is_maximize) and (-loss_change > self.tol)
        if max_improved or min_improved:
            self.best_loss = current_loss.item() if isinstance(current_loss, torch.Tensor) else current_loss
            self.best_epoch = epoch
            self.wait = 1
            self.best_weights = copy.deepcopy(self.trainer.network.state_dict())
        else:
            if self.wait >= self.patience:
                self.stopped_epoch = epoch
                self.trainer._stop_training = True
            self.wait += 1

    def on_train_end(self, logs: Optional[Dict[str, Any]] = None) -> None:
        self.trainer.best_epoch = self.best_epoch
        self.trainer.best_cost = self.best_loss

        if self.best_weights is not None:
            self.trainer.network.load_state_dict(self.best_weights)

        if self.stopped_epoch > 0:
            msg = f"\nEarly stopping occurred at epoch {self.stopped_epoch}"
            msg += f" with best_epoch = {self.best_epoch} and " + f"best_{self.early_stopping_metric} = {round(self.best_loss, 5)}"
            print(msg)
        else:
            msg = (
                f"Stop training because you reached max_epochs = {self.trainer.max_epochs}"
                + f" with best_epoch = {self.best_epoch} and "
                + f"best_{self.early_stopping_metric} = {round(self.best_loss, 5)}"
            )
            print(msg)
        wrn_msg = "Best weights from best epoch are automatically used!"
        warnings.warn(wrn_msg, stacklevel=2)


@dataclass
class History(Callback):
    """Callback that records events into a `History` object.
    This callback is automatically applied to
    every SuperModule.

    Parameters
    ---------
    trainer : Any
        Model class to train
    verbose : int
        Print results every verbose iteration

    """

    trainer: Any
    verbose: int = 1

    def __post_init__(self) -> None:
        super().__init__()
        self.samples_seen: float = 0.0
        self.total_time: float = 0.0

    def on_train_begin(self, logs: Optional[Dict[str, Any]] = None) -> None:
        self.history: Dict[str, List[float]] = {"loss": []}
        self.history.update({"lr": []})
        self.history.update({name: [] for name in self.trainer._metrics_names})
        self.start_time: float = logs["start_time"]
        self.epoch_loss: float = 0.0

    def on_epoch_begin(self, epoch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        self.epoch_metrics: Dict[str, float] = {"loss": 0.0}
        self.samples_seen = 0.0

    def on_epoch_end(self, epoch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        self.epoch_metrics["loss"] = self.epoch_loss
        for metric_name, metric_value in self.epoch_metrics.items():
            self.history[metric_name].append(metric_value)
        if self.verbose == 0:
            return
        if epoch % self.verbose != 0:
            return
        msg = f"epoch {epoch:<3}"
        for metric_name, metric_value in self.epoch_metrics.items():
            if metric_name != "lr":
                msg += f"| {metric_name:<3}: {np.round(metric_value, 5):<8}"
        self.total_time = int(time.time() - self.start_time)
        msg += f"|  {str(datetime.timedelta(seconds=self.total_time)) + 's':<6}"
        print(msg)

    def on_batch_end(self, batch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        batch_size: int = logs["batch_size"]
        self.epoch_loss = (self.samples_seen * self.epoch_loss + batch_size * logs["loss"]) / (self.samples_seen + batch_size)
        self.samples_seen += batch_size

    def __getitem__(self, name: str) -> List[float]:
        return self.history[name]

    def __repr__(self) -> str:
        return str(self.history)

    def __str__(self) -> str:
        return str(self.history)


@dataclass
class LRSchedulerCallback(Callback):
    """Wrapper for most torch scheduler functions.

    Parameters
    ---------
    scheduler_fn : Any
        Torch scheduling class
    optimizer : Any
        Optimizer
    scheduler_params : dict
        Dictionnary containing all parameters for the scheduler_fn
    early_stopping_metric : str
        Metric for early stopping
    is_batch_level : bool (default = False)
        If set to False : lr updates will happen at every epoch
        If set to True : lr updates happen at every batch
        Set this to True for OneCycleLR for example
    """

    scheduler_fn: Any
    optimizer: Any
    scheduler_params: Dict[str, Any]
    early_stopping_metric: str
    is_batch_level: bool = False

    def __post_init__(
        self,
    ) -> None:
        self.is_metric_related: bool = hasattr(self.scheduler_fn, "is_better")
        self.scheduler: Any = self.scheduler_fn(self.optimizer, **self.scheduler_params)
        super().__init__()

    def on_batch_end(self, batch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        if self.is_batch_level:
            self.scheduler.step()
        else:
            pass

    def on_epoch_end(self, epoch: int, logs: Optional[Dict[str, Any]] = None) -> None:
        current_loss = logs.get(self.early_stopping_metric)
        if current_loss is None:
            return
        if self.is_batch_level:
            pass
        else:
            if self.is_metric_related:
                self.scheduler.step(current_loss)
            else:
                self.scheduler.step()
